# Back-ported from Python 3.2.
# Copyright (c) 2001-2012, 2014-2016, Python Software Foundation; All rights reserved.
# Retains the licence of the Python Software Foundation.
# flake8: noqa
"""Redo the builtin repr() (representation) but with limits on most sizes."""

from reprlib import Repr, repr, recursive_repr, __all__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Tools for managing function arguments.

Process function arguments could be messy when a flexible schema is needed.
With this module you can outline parameters schema using a smart way of
processing actual arguments:

A parameter row (see `ParamSchemeRow`:class:), allow several keywords IDs (one
is required used as the final identifier for the actual argument). Also
integer IDs expressing logical order for positional argument passing (negative
values are for right-to-left indexing, like in sequences).  Several values
means several possibilities.

.. versionadded:: 1.8.0

'''


#: The maximum number of positional arguments allowed when calling a function.
MAX_ARG_COUNT = 1024 * 1024    # just any large number

from xotl.tools.symbols import Undefined  # used implicitly for absent default


def issue_9137(args):
    '''Parse arguments for methods, fixing `issue 9137`__ (self ambiguity).

    There are methods that expect 'self' as valid keyword argument, this is
    not possible if this name is used explicitly::

        def update(self, *args, **kwds):
            ...

    To solve this, declare the arguments as ``method_name(*args, **kwds)``,
    and in the function code::

        self, args = issue_9137(args)

    :returns: (self, remainder positional arguments in a tuple)

    .. versionadded:: 1.8.0

    __ https://bugs.python.org/issue9137

    '''
    self = args[0]    # Issue 9137
    args = args[1:]
    return self, args


def check_count(args, low, high=MAX_ARG_COUNT, caller=None):
    '''Check the positional arguments actual count against constrains.

    :param args: The args to check count, normally is a tuple, but an integer
           is directly accepted.

    :param low: Integer expressing the minimum count allowed.

    :param high: Integer expressing the maximum count allowed.

    :param caller: Name of the function issuing the check, its value is used
           only for error reporting.

    .. versionadded:: 1.8.0


    '''
    # TODO: Shouldn't we use the TypeError and ValueError?
    assert isinstance(low, int) and low >= 0
    assert isinstance(high, int) and high >= low
    if isinstance(args, int):
        count = args
        if count < 0:
            msg = "check_count() don't accept a negative argument count: {}"
            raise ValueError(msg.format(count))
    else:
        count = len(args)
    if count < low:
        error = True
        adv = 'exactly' if low == high else 'at least'
        if low == 1:
            aux = '{} one argument'.format(adv)
        else:
            aux = '{} {} arguments'.format(adv, low)
    elif count > high:
        error = True
        if low == high:
            if low == 0:
                aux = 'no arguments'
            elif low == 1:
                aux = 'exactly one argument'
            else:
                aux = 'exactly {} arguments'.format(low)
        elif high == 1:
            aux = 'at most one argument'
        else:
            aux = 'at most {} arguments'.format(high)
    else:
        error = False
    if error:
        if caller:
            name = '{}()'.format(caller)
        else:
            name = 'called function or method'
        raise TypeError('{} takes {} ({} given)'.format(name, aux, count))


def check_default(absent=Undefined):
    '''Get a default value passed as a last excess positional argument.

    :param absent: The value to be used by default if no one is given.
           Defaults to `~xotl.tools.symbols.Undefined`:obj:.

    For example::

        def get(self, name, *default):
            from xotl.tools.params import check_default, Undefined
            if name in self.inner_data:
                return self.inner_data[name]
            elif check_default()(*default) is not Undefined:
                return default[0]
            else:
                raise KeyError(name)

    .. versionadded:: 1.8.0

    '''
    def default(res=absent):
        return res
    return default


def single(args, kwds):
    '''Return a true value only when a unique argument is given.

    When needed, the most suitable result will be wrapped using the
    `~xotl.tools.fp.option.Maybe`:class:\ .

    .. versionadded:: 1.8.0

    '''
    from xotl.tools.fp.option import Just, Wrong, take
    if len(args) == 1 and not kwds:
        res = take(args[0])
        if not res:
            res = Just(res)
        res = Just(res)
    elif not args and len(kwds) == 1:
        res = kwds
    else:
        res = Wrong((args, kwds))
    return res


def pop_keyword_arg(kwargs, names, default=Undefined):
    '''Return the value of a keyword argument.

    :param kwargs: The mapping with passed keyword arguments.

    :param names: Could be a single name, or a collection of names.

    :param default: The default value to return if no value is found.

    .. versionadded:: 1.8.0

    '''
    from xotl.tools.objects import pop_first_of
    if isinstance(names, str):
        names = (names,)
    return pop_first_of(kwargs, *names, default=default)


def pop_keyword_values(kwargs, *names, **options):
    '''Return a list with all keyword argument values.

    :param kwargs: The mapping with passed keyword arguments.

    :param names: Each item will be a definition of keyword argument name to
           retrieve.  Could be a string with a name, or a list of alternatives
           (aliases).

    :keyword default: Keyword only option to define a default value to be used
           in place of not given arguments.  If not given, it is used
           special value `~xotl.tools.symbols.Undefined`:obj:.

    :keyword defaults: A dictionary with default values per argument name. If
           none is given, use `default`.

           .. note:: `defaults` trumps `default`.

           .. warning:: For the case where a single name has several
              alternatives, you may choose any of the alternatives.  If you
              pass several diverging defaults for different alternatives, the
              result is undefined.

    :keyword ignore_error: By default, when there are remaining values in
           `kwargs`, after all names are processed, a `TypeError`:class: is
           raised.  If this keyword only option is True, this function returns
           normally.

    Examples::

      >>> pop_keyword_values({'b': 1}, 'a', 'b')
      [Undefined, 1]

      >>> kwargs = {'a': 1, 'b': 2, 'c': 3}
      >>> try:
      ...     res = pop_keyword_values(kwargs, 'a', 'b')
      ... except TypeError as error:
      ...     res = error
      >>> type(res)
      TypeError

      >>> kwargs = {'a': 1, 'b': 2, 'c': 3}
      >>> options = dict(ignore_error=True, default=None)
      >>> pop_keyword_values(kwargs, 'a', ('B', 'b'), **options)
      [1, 2]

    .. versionadded:: 1.8.3

    '''
    default = options.get('default', Undefined)
    defaults = options.get('defaults', {})
    res = []
    for item in names:
        val = pop_keyword_arg(kwargs, item, default=Undefined)
        if val is Undefined:
            val = pop_keyword_arg(defaults, item, default=default)
        res.append(val)
    if kwargs and not options.get('ignore_error', False):
        msg = 'calling function got unexpected keyword arguments "{}"'
        raise TypeError(msg.format(tuple(kwargs)))
    return res


class ParamManager:
    '''Function parameters parser.

    For example::

      def wraps(*args, **kwargs):
          pm = ParamManager(args, kwargs)
          name = pm(0, 1, 'name', coerce=str)
          wrapped = pm(0, 1, 'wrapped', coerce=valid(callable))
          ...

    When an instance of this class is called (``__call__`` operator), it is
    used the same protocol as when creating an instance of a parameter
    definition row (`ParamSchemeRow`:class:).

    See `ParamScheme`:class: class as another way to define and validate
    schemes for extracting parameter values in a consistent way.


    .. versionadded:: 1.8.0

    '''

    def __init__(self, args, kwds):
        '''Created with actual parameters of a client function.'''
        self.args = args
        self.kwds = kwds
        self.consumed = set()    # consumed identifiers

    def __call__(self, *ids, **options):
        '''Get a parameter value.'''
        from xotl.tools.fp.option import Just, Wrong, none
        # TODO: Change this ``from xotl.tools.values import coercer``
        from xotl.tools.fp.prove.semantic import predicate as coercer
        args, kwds = self.args, self.kwds
        i, res = 0, none
        while isinstance(res, Wrong) and i < len(ids):
            key = ids[i]
            if key in self.consumed:
                pass
            elif isinstance(key, int):
                try:
                    res = args[key]
                except IndexError:
                    pass
            elif key in kwds:
                res = kwds[key]
            if not isinstance(res, Wrong) and 'coerce' in options:
                aux = coercer(options['coerce'])(res)
                res = aux.inner if isinstance(aux, Just) else aux
            if not isinstance(res, Wrong):
                self.consumed.add(key)
                if isinstance(key, int) and key < 0:
                    # consume both, negative and adjusted value
                    key = len(args) + key
                    self.consumed.add(key)
            else:
                i += 1
        if isinstance(res, Wrong):
            if 'default' in options:
                return options['default']
            elif isinstance(res.inner, BaseException):
                raise res.inner
            else:
                raise TypeError('value for "{}" is not found'.format(ids))
        else:
            return res.inner if isinstance(res, Just) else res

    def remainder(self):
        '''Return not consumed values in a mapping.'''
        passed = set(range(len(self.args))) | set(self.kwds)
        ids = passed - self.consumed
        args, kwds = self.args, self.kwds
        return {k: args[k] if isinstance(k, int) else kwds[k] for k in ids}


class ParamSchemeRow:
    '''Scheme row for a  `ParamManager`:class: instance call.

    This class validates identifiers and options at this level; these
    checks are not done in a call to get a parameter value.

    Normally this class is used as part of a full `ParamScheme`:class:
    composition.

    Additionally to the options can be passed to
    `ParamManager.__call__`:meth:', this class can be instanced with:

    :param ids: positional variable number arguments, could be aliases for
           keyword parameter passing, or integers for order (negative values
           are means right-to-left indexing, like in sequences);

    :param key: an identifier to be used when the parameter is only positional
           or when none of the possible keyword aliases must be used as the
           primary-key;

    :param default: keyword argument, value used if the parameter is absent;

    :param coerce: check if a value is valid or not and convert to its
           definitive value; see `xotl.tools.values`:mod: module for more
           information.

    .. versionadded:: 1.8.0

    '''
    __slots__ = ('ids', 'options', '_key')

    def __init__(self, *ids, **options):
        from collections import Counter
        from xotl.tools.fp.option import none
        iskey = lambda s: isinstance(s, str) and s.isidentifier()
        # TODO: Change this ``from xotl.tools.values import coercer``
        from xotl.tools.fp.prove.semantic import predicate as coercer
        aux = {k: c for k, c in Counter(ids).items() if c > 1}
        if aux:
            parts = ['{!r} ({})'.format(k, aux[k]) for k in aux]
            msg = '{}() repeated identifiers: {}'
            raise TypeError(msg.format(type(self).__name__, ', '.join(parts)))
        else:
            def ok(k):
                return iskey(k) or isinstance(k, int)

            bad = [k for k in ids if not ok(k)]
            if bad:
                msg = ('{}() identifiers with wrong type (only int and str '
                       'allowed): {}')
                raise TypeError(msg.format(type(self).__name__, bad))
        key = options.pop('key', none)
        if not (key is none or iskey(key)):
            msg = ('"key" option must be an identifier, "{}" of type "{}" '
                   'given')
            raise TypeError(msg.format(key, type(key).__name__))
        if 'default' in options:
            aux = {'default': options.pop('default')}
        else:
            aux = {}
        if 'coerce' in options:
            aux['coerce'] = coercer(options.pop('coerce'))
        if options:
            msg = '{}(): received invalid keyword parameters: {}'
            raise TypeError(msg.format(type(self).__name__, set(options)))
        self.ids = ids
        self.options = aux
        self._key = key

    def __str__(self):
        parts = [repr(k) for k in self.ids]
        for key, value in self.options.items():
            parts.append('{}={!r}'.format(key, value))
        aux = ', '.join(parts)
        return 'ParamSchemeRow({})'.format(aux)
    __repr__ = __str__

    def __call__(self, *args, **kwds):
        '''Execute a scheme-row using as argument a `ParamManager` instance.

        The concept of `ParamManager`:class: instance argument is a little
        tricky: when a variable number of arguments is used, if only one
        positional and is already an instance of `ParamManager`:class:, it is
        directly used; if two, the first is a `tuple` and the second is a
        `dict`, these are considered the constructor arguments of the new
        instance; otherwise all arguments are used to build the new instance.

        '''
        count = len(args)
        if count == 1 and not kwds and isinstance(args[0], ParamManager):
            manager = args[0]
        else:
            if count == 2 and not kwds:
                a, k = args
                if isinstance(a, tuple) and isinstance(k, dict):
                    args, kwds = a, k
            manager = ParamManager(args, kwds)
        return manager(*self.ids, **self.options)

    @property
    def default(self):
        '''Returned value if parameter value is absent.

        If not defined, special value ``none`` is returned.

        '''
        from xotl.tools.fp.option import none
        return self.options.get('default', none)

    @property
    def key(self):
        '''The primary key for this scheme-row definition.

        This concept is a little tricky (the first string identifier if some
        is given, if not then the first integer).  This definition is useful,
        for example, to return remainder not consumed values after a scheme
        process is completed (see `ParamManager.remainder`:meth: for more
        information).

        '''
        # TODO: calculate the key value in the constructor
        from xotl.tools.fp.option import none
        res = self._key
        if res is none:
            res = next((k for k in self.ids if isinstance(k, str)), None)
            if res is None:
                res = self.ids[0]
            self._key = res
        return res


class ParamScheme:
    '''Full scheme for a  `ParamManager`:class: instance call.

    This class receives a set of `ParamSchemeRow`:class: instances and
    validate them as a whole.

    .. versionadded:: 1.8.0

    '''
    __slots__ = ('rows', 'cache')

    def __init__(self, *rows):
        from xotl.tools.params import check_count
        check_count(len(rows) + 1, 2, caller=type(self).__name__)
        used = set()
        for idx, row in enumerate(rows):
            if isinstance(row, ParamSchemeRow):
                this = {k for k in row.ids if isinstance(k, str)}
                aux = used & this
                if not aux:
                    used |= this
                else:
                    msg = ('{}() repeated keyword identifiers "{}" in '
                           'row {}').format(type(self).__name__, aux, idx)
                    raise ValueError(msg)
        self.rows = rows
        self.cache = None

    def __str__(self):
        # XXX: Use:: ',\n\i'.join(map(str, self))
        aux = ',\n\i'.join(str(row) for row in self)
        return '{}({})'.format(type(self).__name__, aux)

    def __repr__(self):
        return '{}({} rows)'.format(type(self).__name__, len(self))

    def __len__(self):
        '''The defined scheme-rows number.'''
        return len(self.rows)

    def __getitem__(self, idx):
        '''Obtain the scheme-row by a given index.'''
        if isinstance(idx, str):
            cache = self._getcache()
            return cache[idx]
        else:
            return self.rows[idx]

    def __iter__(self):
        '''Iterate over all defined scheme-rows.'''
        return iter(self.rows)

    def __call__(self, args, kwds, strict=True):
        '''Get a mapping with all resulting values.

        If special value 'none' is used as 'default' option in a scheme-row,
        corresponding value isn't returned in the mapping if the parameter
        value is missing.

        '''
        def ok(v):
            from xotl.tools.fp.option import Wrong
            return not isinstance(v, Wrong)

        pm = ParamManager(args, kwds)
        aux = ((row.key, row(pm)) for row in self)
        res = {key: value for key, value in aux if ok(value)}
        rem = pm.remainder()
        if strict:
            if rem:
                msg = ('after a full `{}` process, there are still remainder '
                       'parameters: {}')
                raise TypeError(msg.format(type(self).__name__, set(rem)))
        else:
            res.update(rem)
        return res

    def keys(self):
        '''Partial compatibility with mappings.'''
        return self._getcache().keys()

    def items(self):
        '''Partial compatibility with mappings.'''
        return self._getcache().items()

    @property
    def defaults(self):
        '''Return a mapping with all valid default values.'''
        def ok(v):
            from xotl.tools.fp.option import Wrong
            return not isinstance(v, Wrong)

        aux = ((row.key, row.default) for row in self)
        return {k: d for k, d in aux if ok(d)}

    def _getcache(self):
        if not self.cache:
            self.cache = {row.key: row for row in self}
        return self.cache



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import sys
import importlib
from importlib.abc import MetaPathFinder

XOUTIL_NAMESPACE = 'xoutil.'
XOTL_TOOLS_NS = 'xotl.tools.'


class Hook(MetaPathFinder):
    def find_module(self, full_name, path=None):
        name = self._from_xoutil_to_xotl(full_name)
        if name:
            return self

    def _from_xoutil_to_xotl(self, full_name):
        if full_name.startswith(XOUTIL_NAMESPACE):
            path = full_name[len(XOUTIL_NAMESPACE):]
            return XOTL_TOOLS_NS + path
        else:
            return None

    def load_module(self, full_name):
        result = sys.modules.get(full_name, None)
        if result:
            return result
        modname = self._from_xoutil_to_xotl(full_name)
        if modname:
            result = sys.modules[full_name] = importlib.import_module(modname)
            return result
        else:
            raise ImportError(modname)


sys.meta_path.append(Hook())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Simple tools minimizing dependencies on other modules.

The only used module is Python's standard `re`:mod: module.

'''

from xotl.tools.deprecation import deprecated
from xotl.tools.params import check_default


# TODO: review this
def nameof(obj):
    '''Give the name of an object.

    First try normally named object (those having a ``'__name__'`` attribute);
    then some special classes instances that wraps the name internally are
    checked; then it tests for some objects that are singletons\ [#sing]_;
    finally -as a default tentative- return the type name.

    For example::

      >>> nameof(object)
      'object'

      >>> nameof(lambda x: x)
      '<lambda>'

      >>> singletons = (None, True, False, Ellipsis, NotImplemented)
      >>> [nameof(s) for s in singletons]
      ['None', 'True', 'False', 'Ellipsis', 'NotImplemented']

      >>> nameof(0)
      'int'

    This is a beginning intended to deprecate a "fauna" of several existing
    functions with the same purpose around several modules.

    .. [#sing] In this case an object is considered a signgleton if both of
               its representation strings (``str(obj)`` and ``repr(obj)``)
               match and it is a valid identifier.

    '''
    try:
        return obj.__name__
    except AttributeError:
        if isinstance(obj, (staticmethod, classmethod)):
            return obj.__func__.__name__
        else:    # try for singleton
            import re
            res = str(obj)
            identifier_regex = '(?i)^[_a-z][_a-z0-9]*$'    # TODO: Py3?
            if res == repr(obj) and re.match(identifier_regex, res):
                return res
            else:
                return type(obj).__name__


# TODO: Move all functions in this module to a new place

@deprecated(check_default)
def get_default(args, default=None):
    '''Get a default value passed as last positional argument.

    See `xotl.tools.params.check_default`:func: for a more accurate function
    that replace this.

    .. deprecated:: 1.8.0

    '''
    return check_default(default)(*args)


def args_repr(args, **options):
    '''Format positional arguments to use in exception handling.

    :params args: tuple as obtained in arguments when declared in a function
           as ``*args``.

    :param options: some extra options could be used as excess keyword
           arguments.

           - count: maximum amount of actual parameters to process; after
             reached that amount a tail item is appended with the remainder
             number.  If None is given -the default- the value ``3`` is used.

           - cast: must be a function to convert the value into the
             representation; when no value is given -the default- it's assumed
             ``λ arg: type(arg).__name__``.

           - item_format: the format for each argument type, if no value is
             given the value "{}" is used.  Each item is obtained as with the
             "cast" function.

           - tail_format: a format string for the tail with the remainder (not
             processed arguments) specification; must contain a positional
             (like "{}") format specifier if obtaining the remainder count is
             desired; when no value is given -the default- the value "..." is
             used; another valid example could be "and {} more".

           - joiner: could be a function or a string to join all parts
             including the tail; if no value is given -the default- the value
             ", " is used (thed equivalent to ``', '.join``).

    For example::

      >>> args_repr((1, 2.0, "3", {}))
      'int, float, str, ...'

    '''
    count = options.get('count', 3)
    cast = options.get('cast', lambda arg: type(arg).__name__)
    item_format = options.get('item_format', '{}')
    tail_format = options.get('tail_format', '...')
    joiner = options.get('joiner', ', ')
    if isinstance(joiner, str):
        joiner = str(joiner).join
    parts = []
    i = 0
    while i < min(count, len(args)):
        parts.append(item_format.format(cast(args[i])))
        i += 1
    rem = len(args) - i
    if rem > 0:
        parts.append(tail_format.format(rem))
    return joiner(parts)


def kwargs_repr(kwargs, **options):
    '''Format positional arguments to use in exception handling.

    :params kwargs: dict as obtained in arguments when declared as
           ``**kwargs``.

    :param options: some extra options are used in this function.

           - count: maximum amount of actual parameters to process; after
             reached that amount a tail item is appended with the remainder
             number.  If None is given -the default- the value ``3`` is used.

           - cast: must be a function to convert the value into the
             representation; when no value is given -the default- it's assumed
             ``λ arg: type(arg).__name__``.

           - item_format: the format for each argument type, if no value is
             given the value "{}:{}" is used.  Each item value is
             obtained as with the "cast" function.

           - tail_format: a format string for the tail with the remainder (not
             processed arguments) specification; must contain a positional
             (like "{}") format specifier if obtaining the remainder count is
             desired; when no value is given -the default- the value "..." is
             used; another valid example could be "and {} more".

           - joiner: could be a function or a string to join all parts
             including the tail; if no value is given -the default- the value
             ", " is used (thed equivalent to ``', '.join``).

    For example::

      >>> kwargs_repr({'x': 1, 'y': 2.0, 'z': '3', 'zz': {}})
      'x:int, y:float, z:str, ...'

    '''
    count = options.get('count', 3)
    cast = options.get('cast', lambda arg: type(arg).__name__)
    item_format = options.get('item_format', '{}:{}')
    tail_format = options.get('tail_format', '...')
    joiner = options.get('joiner', ', ')
    if isinstance(joiner, str):
        joiner = str(joiner).join
    parts = []
    keys = list(kwargs)
    keys.sort()
    i = 0
    while i < min(count, len(keys)):
        key = keys[i]
        value = kwargs[key]
        parts.append(item_format.format(key, cast(value)))
        i += 1
    rem = len(keys) - i
    if rem > 0:
        parts.append(tail_format.format(rem))
    return joiner(parts)


def both_args_repr(args, kwargs, **options):
    '''Combine both argument kind representations.

    Both kinds are: positional (see `args_repr`:func:) and keyword (see
    `kwargs_repr`:func:).

    For example::

      >>> both_args_repr((1, 2.0, "3"), {'x': 1, 'y': 2.0, 'z': '3'})
      'int, float, str, x:int, y:float, z:str'

    '''
    joiner = options.get('joiner', ', ')
    if isinstance(joiner, str):
        joiner = str(joiner).join
    items = (args, args_repr), (kwargs, kwargs_repr)
    parts = [res for res in (fn(aux, **options) for aux, fn in items) if res]
    return joiner(parts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

# This module is a modified copy of gevent.local.  We simply modify it to
# allow absence of greenlets, and fallback to thread isolation in this case.
#
# This decouples xotl.tools.context from gevent and allows to use the
# greenlets if available.

# WARNING: We removed the greenlet protection gevent.local does while
# initializing a subclass of `local`.  Instead we simply provide protection at
# the thread level, so sub-classes of `local` MUST NOT switch greenlets.
from threading import RLock


# since each thread has its own greenlet we can just use those as identifiers
# for the context.  If greenlets are not available we fall back to the
# current thread ident depending on where it is.
try:
    from greenlet import getcurrent
except ImportError:
    from threading import current_thread as getcurrent


from weakref import WeakKeyDictionary
from copy import copy


import sys
PYPY = hasattr(sys, 'pypy_version_info')

__all__ = ["local"]


class _localbase:
    __slots__ = '_local__args', '_local__lock', '_local__dicts'

    def __new__(cls, *args, **kw):
        self = object.__new__(cls)
        object.__setattr__(self, '_local__args', (args, kw))
        object.__setattr__(self, '_local__lock', RLock())
        dicts = WeakKeyDictionary()
        object.__setattr__(self, '_local__dicts', dicts)

        if args or kw:
            clsi, obji = cls.__init__, object.__init__
            if (PYPY and clsi == obji) or (not PYPY and clsi is obji):
                raise TypeError("Initialization arguments are not supported")

        # We need to create the greenlet dict in anticipation of
        # __init__ being called, to make sure we don't call it again ourselves.
        dict = object.__getattribute__(self, '__dict__')
        dicts[getcurrent()] = dict
        return self


def _init_locals(self):
    d = {}
    dicts = object.__getattribute__(self, '_local__dicts')
    dicts[getcurrent()] = d
    object.__setattr__(self, '__dict__', d)

    # we have a new instance dict, so call out __init__ if we have one
    cls = type(self)
    if cls.__init__ is not object.__init__:
        args, kw = object.__getattribute__(self, '_local__args')
        cls.__init__(self, *args, **kw)


class local(_localbase):
    '''Greenlet-local data.'''
    def __getattribute__(self, name):
        d = object.__getattribute__(self, '_local__dicts').get(getcurrent())
        if d is None:
            # it's OK to acquire the lock here and not earlier, because the
            # above code won't switch out however, subclassed __init__ might
            # switch, so we do need to acquire the lock here
            lock = object.__getattribute__(self, '_local__lock')
            lock.acquire()
            try:
                _init_locals(self)
                return object.__getattribute__(self, name)
            finally:
                lock.release()
        else:
            object.__setattr__(self, '__dict__', d)
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name == '__dict__':
            clsname = self.__class__.__name__
            raise AttributeError(
                "%r object attribute '__dict__' is read-only" % clsname
            )
        d = object.__getattribute__(self, '_local__dicts').get(getcurrent())
        if d is None:
            lock = object.__getattribute__(self, '_local__lock')
            lock.acquire()
            try:
                _init_locals(self)
                return object.__setattr__(self, name, value)
            finally:
                lock.release()
        else:
            object.__setattr__(self, '__dict__', d)
            return object.__setattr__(self, name, value)

    def __delattr__(self, name):
        if name == '__dict__':
            clsname = self.__class__.__name__
            raise AttributeError(
                "%r object attribute '__dict__' is read-only" % clsname
            )
        d = object.__getattribute__(self, '_local__dicts').get(getcurrent())
        if d is None:
            lock = object.__getattribute__(self, '_local__lock')
            lock.acquire()
            try:
                _init_locals(self)
                return object.__delattr__(self, name)
            finally:
                lock.release()
        else:
            object.__setattr__(self, '__dict__', d)
            return object.__delattr__(self, name)

    def __copy__(self):
        currentId = getcurrent()
        d = object.__getattribute__(self, '_local__dicts').get(currentId)
        duplicate = copy(d)

        cls = type(self)
        if cls.__init__ is not object.__init__:
            args, kw = object.__getattribute__(self, '_local__args')
            instance = cls(*args, **kw)
        else:
            instance = cls()

        object.__setattr__(instance, '_local__dicts', {
            currentId: duplicate
        })

        return instance



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

VERSION = '2.1.0'


def dev_tag_installed():
    import pkg_resources
    try:
        dist = pkg_resources.get_distribution('xotl.tools')
        full_version = dist.version
        # FIX: Below line is not working anymore
        base = dist.parsed_version.base_version
        return full_version[len(base):]
    except Exception:
        return None


RELEASE_TAG = dev_tag_installed() or ''


def safe_int(x):
    try:
        return int(x)
    except ValueError:
        return x


# I won't put the release tag in the version_info tuple.  Since PEP440 is on
# the way.
VERSION_INFO = tuple(safe_int(x) for x in VERSION.split('.'))


del safe_int



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Collection of disparate utilities.

`xotl.tools` is essentially an extension to the Python's standard library, it
does not make up a full framework, but it's very useful to be used from a
diversity of scenarios.

'''
import sys
from .modules import customize
from .deprecation import DeprecatedImportDescriptor

customize(sys.modules[__name__], custom_attrs=dict(
    Unset=DeprecatedImportDescriptor('xotl.tools.symbols.Unset'),
    Undefined=DeprecatedImportDescriptor('xotl.tools.symbols.Undefined'),
    Ignored=DeprecatedImportDescriptor('xotl.tools.symbols.Ignored'),
    Invalid=DeprecatedImportDescriptor('xotl.tools.symbols.Invalid'),
))

del customize, sys, DeprecatedImportDescriptor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Special logical values like Unset, Undefined, Ignored, Invalid, ...

All values only could be `True` or `False` but are intended in places where
`None` is expected to be a valid value or for special Boolean formats.

'''


SYMBOL = 'symbol'
BOOLEAN = 'boolean'

TIMEOUT = 2.0


class MetaSymbol(type):
    '''Meta-class for symbol types.'''
    def __new__(cls, name, bases, ns):
        if ns['__module__'] == __name__ or name not in {SYMBOL, BOOLEAN}:
            self = super().__new__(cls, name, bases, ns)
            if name == SYMBOL:
                self._instances = {str(v): v for v in (False, True)}
            return self
        else:
            raise TypeError('invalid class "{}" declared outside of "{}" '
                            'module'.format(name, __name__))

    def __instancecheck__(self, instance):
        '''Override for isinstance(instance, self).'''
        if instance is False or instance is True:
            return True
        else:
            return super().__instancecheck__(instance)

    def __subclasscheck__(self, subclass):
        '''Override for issubclass(subclass, self).'''
        if subclass is bool:
            return True
        else:
            return super().__subclasscheck__(subclass)

    def nameof(self, s):
        '''Get the name of a symbol instance (`s`).'''
        items = self._instances.items()
        return next((name for name, value in items if value is s), None)

    def parse(self, name):
        '''Returns instance from a string.

        Standard Python Boolean values are parsed too.

        '''
        if '#' in name:    # Remove comment
            name = name.split('#')[0].strip()
        res = self._instances.get(name, None)
        if res is not None:
            if isinstance(res, self):
                return res
            else:
                msg = 'invalid parsed value "{}" of type "{}"; must be "{}"'
                rtn, sn = type(res).__name__, self.__name__
                raise TypeError(msg.format(res, rtn, sn))
        else:
            msg = 'name "{}" is not defined'
            raise NameError(msg.format(name))


class symbol(int, metaclass=MetaSymbol):
    '''Instances are custom symbols.

    Symbol instances identify uniquely a semantic concept by its name.  Each
    one has an ordinal value associated.

    For example::

      >>> ONE2MANY = symbol('ONE2MANY')
      >>> ONE_TO_MANY = symbol('ONE2MANY')

      >>> ONE_TO_MANY is ONE2MANY
      True

    '''
    __slots__ = ()

    def __new__(cls, name, value=None):
        '''Get or create a new symbol instance.

        :param name: String representing the internal name.  `Symbol`:class:
               instances are unique (singletons) in the context of this
               argument.  ``#`` and spaces are invalid characters to allow
               comments.

        :param value: Any value compatible with Python `bool` or `int` types.
               `None` is used as a special value to create a value using the
               name hash.

        '''
        from sys import intern as unique
        name = unique(name)
        if name:
            if value is None:
                value = hash(name)
            res = cls._instances.get(name)
            if res is None:    # Create the new instance
                if isinstance(value, int):
                    res = super().__new__(cls, value)
                    cls._instances[name] = res
                else:
                    msg = ('instancing "{}" with name "{}" and incorrect '
                           'value "{}" of type "{}"')
                    cn, vt = cls.__name__, type(value).__name__
                    raise TypeError(msg.format(cn, name, value, vt))
            elif res != value:    # Check existing instance
                msg = 'value "{}" mismatch for existing instance: "{}"'
                raise ValueError(msg.format(value, name))
            return res
        else:
            raise ValueError('name must be a valid non empty string')

    def __init__(self, *args, **kwds):
        pass

    def __repr__(self):
        return symbol.nameof(self)

    __str__ = __repr__


class boolean(symbol):
    '''Instances are custom logical values (`True` or `False`).

    Special symbols allowing only logical (False or True) values.

    For example::

      >>> true = boolean('true', True)
      >>> false = boolean('false')
      >>> none = boolean('false')
      >>> unset = boolean('unset')

      >>> class X:
      ...      attr = None

      >>> getattr(X(), 'attr') is not None
      False

      >>> getattr(X(), 'attr', false) is not false
      True

      >>> none is false
      True

      >>> false == False
      True

      >>> false == unset
      True

      >>> false is unset
      False

      >>> true == True
      True

    '''
    __slots__ = ()

    def __new__(cls, name, value=False):
        '''Get or create a new symbol instance.

        See `~Symbol.__new__`:meth: for information about parameters.
        '''
        return super().__new__(cls, name, bool(value))


# --- Special singleton values ---

#: False value, mainly for function parameter definitions, where None could
#: be a valid value.
Unset = boolean('Unset')

#: False value for local scope use or where ``Unset`` could be a valid value
Undefined = boolean('Undefined')

#: To be used in arguments that are currently ignored because they are being
#: deprecated.  The only valid reason to use `Ignored` is to signal ignored
#: arguments in method's/function's signature
Ignored = boolean('Ignored')

#: To be used in functions resulting in a fail where False could be a valid
#: value.
Invalid = boolean('Invalid')

#: To be used as a mark for current context as a mechanism of comfort.
This = boolean('This', True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Extensions to Python's `collections` module.

You may use it as drop-in replacement of `collections`.  Although we don't
document all items here.  Refer to `collections`:mod: documentation.

'''

from collections import *    # noqa
try:
    from collections.abc import *  # noqa
    from collections.abc import (
        Set,
        Iterable,
        Sized,
        Container,
        MutableSet,
        Mapping,
        MutableMapping
    )
except ImportError:
    from collections import (
        Set,
        Iterable,
        Sized,
        Container,
        MutableSet,
        Mapping,
        MutableMapping
    )


import collections as _stdlib    # noqa
from collections import (_itemgetter, _heapq, _chain, _repeat,    # noqa
                         _starmap, _count_elements)

from xotl.tools.deprecation import deprecated    # noqa
from xotl.tools.symbols import Unset    # noqa
from xotl.tools.objects import SafeDataItem as safe    # noqa
from xotl.tools.reprlib import recursive_repr    # noqa


class safe_dict_iter(tuple):
    '''Iterate a dictionary in a safe way.

    This is useful when a dictionary can be modified while iterating on it,
    for example::

      >>> d = {1: 2, 3: 4, 5: 6}
      >>> di = safe_dict_iter(d)

      >>> for k, v in di.items():
      ...     d[v] = k

      >>> [(k, v) for (k, v) in di.items()]
      [(1, 2), (3, 4), (5, 6)]

      >>> del d[1]

      >>> [(k, v) for (k, v) in di.items()]
      [(3, 4), (5, 6)]

      >>> [k for k in di]
      [3, 5]

    '''

    def __new__(cls, mapping):
        self = super().__new__(cls, mapping)
        self._mapping = mapping
        return self

    def __str__(self):
        cls_name = type(self).__name__
        res = str(', ').join(str(i) for i in self)
        return str('{}({})').format(cls_name, res)
    __repr__ = __str__

    def __len__(self):
        return sum(1 for key in self)

    def __contains__(self, key):
        res = super().__contains__(key)
        return res and key in self.mapping

    def __nonzero__(self):
        return bool(len(self))
    __bool__ = __nonzero__

    def __iter__(self):
        for key in super().__iter__():
            if key in self._mapping:
                yield key
    keys = __iter__

    def values(self):
        for key in self:
            if key in self._mapping:
                yield self._mapping[key]

    def items(self):
        for key in self:
            if key in self._mapping:
                yield (key, self._mapping[key])


class defaultdict(_stdlib.defaultdict):
    '''A hack for ``collections.defaultdict`` that passes the key and a copy of
    self as a plain dict (to avoid infinity recursion) to the callable.

    Examples::

        >>> from xotl.tools.future.collections import defaultdict
        >>> d = defaultdict(lambda key, d: 'a')
        >>> d['abc']
        'a'

    Since the second parameter is actually a dict-copy, you may (naively) do
    the following::

        >>> d = defaultdict(lambda k, d: d[k])
        >>> d['abc']
        Traceback (most recent call last):
            ...
        KeyError: 'abc'


    You may use this class as a drop-in replacement for
    ``collections.defaultdict``::

        >>> d = defaultdict(lambda: 1)
        >>> d['abc']
        1

    '''
    def __missing__(self, key):
        if self.default_factory is not None:
            try:
                return self.default_factory(key, dict(self))
            except TypeError:
                # This is the error when the arguments are not expected.
                return super().__missing__(key)
        else:
            raise KeyError(key)


class OpenDictMixin:
    '''A mixin for mappings implementation that expose keys as attributes.

    For example:

      >>> from xotl.tools.objects import SafeDataItem as safe
      >>> class MyOpenDict(OpenDictMixin, dict):
      ...     __slots__ = safe.slot(OpenDictMixin.__cache_name__, dict)

      >>> d = MyOpenDict({'es': 'spanish'})
      >>> d.es
      'spanish'

      >>> d['es'] = 'espanol'
      >>> d.es
      'espanol'

    When setting or deleting an attribute, the attribute name is regarded as
    key in the mapping if neither of the following condition holds:

    - The name is a `slot`.

    - The object has a ``__dict__`` attribute and the name is key there.

    This mixin defines the following features that can be redefined:

    ``_key2identifier``

        Protected method, receives a key as argument and must return a valid
        identifier that is used instead the key as an extended attribute.

    ``__cache_name__``

        Inner field to store a cached mapping between actual keys and
        calculated attribute names.  The field must be always implemented as a
        `SafeDataItem` descriptor and must be of type `dict`.  There are two
        ways of implementing this:

        - As a slot.  The first time of this implementation is an example.
          Don't forget to pass the second parameter with the constructor
          `dict`.

        - As a normal descriptor::

            >>> from xotl.tools.objects import SafeDataItem as safe
            >>> class MyOpenDict(OpenDictMixin, dict):
            ...     safe(OpenDictMixin.__cache_name__, dict)

    Classes or Mixins that can be integrated with `dict` by inheritance must
    not have a `__slots__` definition.  Because of that, this mixin must not
    declare any slot.  If needed, it must be declared explicitly in customized
    classed like in the example in the first part of this documentation or in
    the definition of `opendict` class.

    '''
    __cache_name__ = str('_inverted_cache')

    def __dir__(self):
        '''Return normal "dir" plus valid keys as attributes.'''
        # TODO: Check if super must be called if defined
        from xotl.tools.objects import fulldir
        return list(set(~self) | fulldir(self))

    def __getattr__(self, name):
        from xotl.tools.future.inspect import get_attr_value
        _mark = object()
        res = get_attr_value(self, name, _mark)
        if res is not _mark:
            return res
        else:
            key = (~self).get(name)
            if key:
                return self[key]
            else:
                msg = "'%s' object has no attribute '%s'"
                raise AttributeError(msg % (type(self).__name__, name))

    def __setattr__(self, name, value):
        cls = type(self)
        desc = getattr(cls, name, None)
        if desc is not None:    # Prioritize descriptors
            try:
                desc.__set__(self, value)
            except Exception:
                pass
        key = (~self).get(name)
        if key:
            self[key] = value
        else:
            super().__setattr__(name, value)

    def __delattr__(self, name):
        key = (~self).get(name)
        if key:
            del self[key]
        else:
            super().__delattr__(name)

    def __invert__(self):
        '''Return an inverted mapping between key and attribute names.

        Keys of the resulting dictionary are identifiers for attribute names
        and values are original key names.

        Class attribute "__separators__" are used to calculate it and is
        cached in '_inverted_cache slot safe variable.

        Several keys could have the same identifier, only one will be valid and
        used.

        To obtain this mapping you can use as the unary operator "~".

        '''
        KEY_LENGTH = 'length'
        KEY_MAPPING = 'mapping'
        cache = self._cache
        cached_length = cache.setdefault(KEY_LENGTH, 0)
        length = len(self)
        if cached_length != length:
            cache[KEY_LENGTH] = length
            aux = ((self._key2identifier(k), k) for k in self)
            res = {key: attr for key, attr in aux if key}
            cache[KEY_MAPPING] = res
        else:
            res = cache.get(KEY_MAPPING)
            if res is None:
                assert cached_length == 0
                res = {}
                cache[KEY_MAPPING] = res
        return res

    @property
    def _cache(self):
        from xotl.tools.future.inspect import get_attr_value
        try:
            return get_attr_value(self, type(self).__cache_name__)
        except AttributeError:
            res = setattr(self, type(self).__cache_name__, dict())
            return res

    @staticmethod
    def _key2identifier(key):
        '''Convert keys to valid identifiers.

        This method could be redefined in sub-classes to change this feature.
        This function must return a valid identifier or None if the conversion
        is not possible.

        '''
        # TODO: Improve this in order to obtain a full-mapping.  For example,
        # the corresponding attribute names for the keys ``'-x-y'`` and
        # ``'x-y'`` are the same, in that case only one will be returning.
        from xotl.tools.keywords import suffix_kwd
        from xotl.tools.string import slugify
        from xotl.tools.validators import is_valid_identifier
        res = key if is_valid_identifier(key) else slugify(key, '_')
        return suffix_kwd(res)


class SmartDictMixin:
    '''A mixin that extends the `update` method of dictionaries

    Standard method allow only one positional argument, this allow several.

    Note on using mixins in Python: method resolution order is calculated in
    the order of inheritance, if a mixin is defined to overwrite behavior
    already existent, use first that classes with it. See `SmartDict`:class:
    below.

    '''
    def update(*args, **kwds):
        '''Update this dict from a set of iterables `args` and keyword values
        `kwargs`.

        Each positional argument could be:

        - another mapping (any object implementing "keys" and
          `~object.__getitem__`:meth: methods.

        - an iterable of (key, value) pairs.

        '''
        from xotl.tools.params import issue_9137
        self, args = issue_9137(args)
        for key, value in smart_iter_items(*args, **kwds):
            self[key] = value

    # TODO: Include new argument ``full=True`` to also search in string
    #       values.  Maybe this kind of feature will be better in a function
    #       instead a method.
    def search(self, pattern):
        '''Return new mapping with items which key match a `pattern` regexp.

        This function always tries to return a valid new mapping of the same
        type of the caller instance.  If the constructor of corresponding type
        can't be called without arguments, then look up for a class
        variable named `__search_result_type__` or return a standard
        Python dictionary if not found.

        '''
        from re import compile
        regexp = compile(pattern)
        cls = type(self)
        try:
            res = cls()
        except Exception:
            from xotl.tools.future.inspect import get_attr_value
            creator = get_attr_value(cls, '__search_result_type__', None)
            res = creator() if creator else {}
        for key in self:
            if regexp.search(key):
                res[key] = self[key]
        return res


class SmartDict(SmartDictMixin, dict):
    '''A "smart" dictionary that can receive a wide variety of arguments.

    See `SmartDictMixin.update`:meth: and :meth:`SmartDictMixin.search`.

    '''
    def __init__(*args, **kwargs):
        from xotl.tools.params import issue_9137
        self, args = issue_9137(args)
        super().__init__()
        self.update(*args, **kwargs)


class opendict(OpenDictMixin, dict):
    '''A dictionary implementation that mirrors its keys as attributes.

    For example::

      >>> d = opendict(es='spanish')
      >>> d.es
      'spanish'

      >>> d['es'] = 'espanol'
      >>> d.es
      'espanol'

    Setting attributes not already included  *does not* makes them keys::

      >>> d.en = 'English'
      >>> set(d)
      {'es'}

    '''
    __slots__ = safe.slot(OpenDictMixin.__cache_name__, dict)

    @classmethod
    def from_enum(cls, enumclass):
        '''Creates an opendict from an enumeration class.

        If `enumclass` lacks the ``__members__`` dictionary, take the
        ``__dict__`` of the class disregarding the keys that cannot be `public
        identifiers
        <xotl.tools.validators.identifiers.is_valid_public_identifier>`:func:.
        If `enumclass` has the ``__members__`` attribute this is the same as
        ``opendict(enumclass.__members__)``.

        Example:

        .. code-block:: python

           >>> from xotl.tools.future.collections import opendict
           >>> @opendict.from_enum
           >>> class Foo:
           ...    x = 1
           ...    _y = 2

           >>> type(Foo) is opendict
           True

           >>> dict(Foo)
           {'x': 1}

        '''
        from xotl.tools.symbols import Unset
        from xotl.tools.validators.identifiers \
            import is_valid_public_identifier
        members = getattr(enumclass, '__members__', Unset)
        if members is Unset:
            members = {
                k: v
                for k, v in enumclass.__dict__.items()
                if is_valid_public_identifier(k)
            }
        return cls(members)


class codedict(OpenDictMixin, dict):
    '''A dictionary implementation that evaluate keys as Python expressions.

    This is also a open dict (see `OpenDictMixin`:class: for more info).

    Example:

      >>> cd = codedict(x=1, y=2, z=3.0)
      >>> '{_[x + y]} is 3 --  {_[x + z]} is 4.0'.format(_=cd)
      '3 is 3 --  4.0 is 4.0'

    It supports the right shift (``>>``) operator as a format operand (using
    ``_`` as the special name for the code dict):

      >>> cd >> '{_[x + y]} is 3 --  {_[x + z]} is 4.0 -- {x} is 1'
      '3 is 3 --  4.0 is 4.0 -- 1 is 1'

    It also implements the left shift (``<<``) operator:

      >>> '{_[x + y]} is 3 --  {_[x + z]} is 4.0 -- {x} is 1' << cd
      '3 is 3 --  4.0 is 4.0 -- 1 is 1'

    .. versionadded:: 1.8.3

    '''
    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            if isinstance(key, str):
                return eval(key, dict(self))
            else:
                raise

    def __rshift__(self, arg):
        return arg.format(_=self, **self)

    __rlshift__ = __rshift__


class StackedDict(OpenDictMixin, SmartDictMixin, MutableMapping):
    '''A multi-level mapping.

    A level is entered by using the `push`:meth: and is leaved by calling
    `pop`:meth:.

    The property `level`:attr: returns the actual number of levels.

    When accessing keys they are searched from the latest level "upwards", if
    such a key does not exists in any level a KeyError is raised.

    Deleting a key only works in the *current level*; if it's not defined there
    a KeyError is raised. This means that you can't delete keys from the upper
    levels without `popping <pop>`:func:.

    Setting the value for key, sets it in the current level.

    .. versionchanged:: 1.5.2 Based on the newly introduced `ChainMap`:class:.

    '''
    __slots__ = (safe.slot('inner', ChainMap),
                 safe.slot(OpenDictMixin.__cache_name__, dict))

    def __init__(*args, **kwargs):
        # Each data item is stored as {key: {level: value, ...}}
        from xotl.tools.params import issue_9137
        self, args = issue_9137(args)
        self.update(*args, **kwargs)

    @property
    def level(self):
        '''Return the current level number.

        The first level is 0.  Calling `push`:meth: increases the current
        level (and returns it), while calling `pop`:meth: decreases the
        current level (if possible).

        '''
        return len(self.inner.maps) - 1

    def push_level(self, *args, **kwargs):
        '''Pushes a whole new level to the stacked dict.

        :param args: Several mappings from which the new level will be
                     initialled filled.

        :param kwargs: Values to fill the new level.

        :returns: The pushed `level`:attr: number.

        '''
        self.inner = self.inner.new_child()
        self.update(*args, **kwargs)
        return self.level

    @deprecated(push_level)
    def push(self, *args, **kwargs):
        '''Don't use thid method, use new `push_level`:meth: instead.'''
        return self.push_level(*args, **kwargs)

    def pop(self, *args):
        '''Remove this, always use original `MutableMapping.pop`:meth:.

        If none arguments are given, `pop_level`:meth: is called and a
        deprecation warning is printed in `sys.stderr` the first time.  If one
        or two arguments are given, those are interpreted as (key, default)
        values of the super class `pop`:meth:.

        '''
        if len(args) == 0:
            cls = type(self)
            if not hasattr(cls, '_bad_pop_called'):
                import warnings
                setattr(cls, '_bad_pop_called', True)
                msg = ('Use `pop` method without parameters is deprecated, '
                       'use `pop_level` instead')
                warnings.warn(msg, stacklevel=2)
            return self.pop_level()
        else:
            return super().pop(*args)

    def pop_level(self):
        '''Pops the last pushed level and returns the whole level.

        If there are no levels in the stacked dict, a TypeError is raised.

        :returns:  A dict containing the poped level.

        '''
        if self.level > 0:
            stack = self.inner
            res = stack.maps[0]
            self.inner = stack.parents
            return res
        else:
            raise TypeError('Cannot pop from StackedDict without any levels')

    def peek(self):
        '''Peeks the top level of the stack.

        Returns a copy of the top-most level without any of the keys from
        lower levels.

        Example::

           >>> sdict = StackedDict(a=1, b=2)
           >>> sdict.push(c=3)  # it returns the level...
           1
           >>> sdict.peek()
           {'c': 3}

        '''
        return dict(self.inner.maps[0])

    def __str__(self):
        # TODO: Optimize
        return str(dict(self))

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, str(self))

    def __len__(self):
        return len(self.inner)

    def __iter__(self):
        return iter(self.inner)

    def __getitem__(self, key):
        return self.inner[key]

    def __setitem__(self, key, value):
        self.inner[key] = value

    def __delitem__(self, key):
        del self.inner[key]


class RankedDict(SmartDictMixin, dict):
    '''Mapping that remembers modification order.

    Differences with `OrderedDict`:class: are:

    - Can be ranked (change precedence order) at any time; see `rank`:meth:
      and `swap_ranks`:meth: methods for more information.

    - Based in modification, not in insertion order.

    - Keeps the standard semantics of Python for `popitem`:meth: method
      returning a random pair when called without parameters.

    '''

    def __init__(*args, **kwds):
        '''Initialize a ranked dictionary.

        The signature is the same as regular dictionaries, but keyword
        arguments are not recommended because their insertion order is
        arbitrary.

        Use `rank`:meth: to change the precedence order in any moment.

        '''
        from xotl.tools.params import issue_9137
        self, args = issue_9137(args)
        # Ensure direct calls to ``__init__`` don't clear previous contents
        try:
            self._ranks
        except AttributeError:
            self._ranks = []
        self.update(*args, **kwds)

    def rank(self, *keys):
        '''Arrange mapping keys into a systematic precedence order.

        :param keys: Variable number of key values, given are ordered in the
               highest precedence levels (0 is the most priority).  Not given
               keys are added at the end in the its current order.

        '''
        if keys:
            ranks = []
            for key in keys:
                if key in self:
                    ranks.append(key)
                else:
                    raise KeyError('{}'.format(key))
            aux = set(keys)
            for key in self:
                if key not in aux:
                    ranks.append(key)
            self._ranks = ranks

    def swap_ranks(self, *args, **kwds):
        '''Exchange ranks of given keys.

        :param args: Each item must be a pair of keys to exchange the order of
               precedence.

        :param kwds: Every keyword argument will have a pair to exchange
               (name, value).

        '''
        from xotl.tools.params import check_count
        check_count(len(args) + len(kwds) + 1, 2, caller='swap_ranks')
        for key1, key2 in args:
            self._swap_ranks(key1, key2)
        for key1 in kwds:
            key2 = kwds[key1]
            self._swap_ranks(key1, key2)

    def move_to_end(self, key, last=True):
        '''Move an existing element to the end.

        Or move it to the beginning if ``last==False``.

        Raises KeyError if the element does not exist.  When ``last==True``,
        acts like a fast version of ``self[key] = self.pop(key)``.

        .. note:: This method is kept for compatibility with
                  `OrderedDict`:class:\ .  Last example using
                  ``self.pop(key)`` works well in both, in OD and this class,
                  but here in `RankedDict`:class: it's semantically equivalent
                  to ``self[key] = self[key]``.

        '''
        try:
            ranks = self._ranks
            if last:
                if key != ranks[-1]:
                    ranks.remove(key)
                    ranks.append(key)
            else:
                if key != ranks[0]:
                    ranks.remove(key)
                    ranks.insert(0, key)
        except ValueError:
            raise KeyError(key)

    def _swap_ranks(self, key1, key2):
        '''Protected method to swap a pair of ranks.'''
        if key1 in self and key2 in self:
            ranks = self._ranks
            idx1, idx2 = ranks.index(key1), ranks.index(key2)
            if idx1 != idx2:
                aux = ranks[idx1]
                ranks[idx1] = ranks[idx2]
                ranks[idx2] = aux
        else:
            raise KeyError('{!r} and/or {!r}'.format(key1, key2))

    def __setitem__(self, key, value):
        '''rd.__setitem__(i, y) <==> rd[i]=y'''
        ranks = self._ranks
        if key in self:
            if ranks[-1] != key:
                ranks.remove(key)
                ranks.append(key)
        else:
            ranks.append(key)
        super().__setitem__(key, value)

    def __delitem__(self, key):
        '''rd.__delitem__(y) <==> del rd[y]'''
        super().__delitem__(key)
        self._ranks.remove(key)

    def __iter__(self):
        '''rd.__iter__() <==> iter(rd)'''
        return iter(self._ranks)

    def __reversed__(self):
        '''rd.__reversed__() <==> reversed(rd)'''
        return reversed(self._ranks)

    def clear(self):
        '''rd.clear() -> None.  Remove all items from rd.'''
        super().clear()
        self._ranks = []

    def popitem(self, index=None):
        '''rd.popitem([index]) -> (key, value), return and remove a pair.

        :param index: Position of pair to return (default last).  This method
               isn't have the same semantic as in `OrderedDict`:class: one,
               none the defined in method with the same name in standard
               Python mappings; here is similar to `~list.pop`:meth:.

        '''
        if self:
            if index is None or index is True:
                index = -1
            key = self._ranks.pop(index)
            return key, super().pop(key)
        else:
            raise KeyError('popitem(): dictionary is empty')

    def __sizeof__(self):
        '''D.__sizeof__() -> size of D in memory, in bytes.

        .. note:: Why ``sys.getsizeof(obj)`` doesn't return the same result as
                  ``obj.__sizeof__()``?

        '''
        return super().__sizeof__() + self._ranks.__sizeof__()

    def keys(self):
        '''D.keys() -> an object providing a view on D's keys.'''
        return self.__iter__()

    def values(self):
        '''D.values() -> an object providing a view on D's values.'''
        for key in self:
            yield self[key]

    def items(self):
        '''D.items() -> an object providing a view on D's items.'''
        for key in self:
            yield (key, self[key])

    def __eq__(self, other):
        '''rd.__eq__(y) <==> rd==y.

        Comparison to another `RankedDict`:class: instance is order-sensitive
        while comparison to a regular mapping is order-insensitive.

        '''
        res = super().__eq__(other)
        if res:
            if isinstance(other, RankedDict):
                return self._ranks == other._ranks
            elif isinstance(other, OrderedDict):
                return self._ranks == list(other)
            else:
                return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def pop(self, key, *args):
        '''rd.pop(k[,d]) -> v

        Remove specified key and return the corresponding value.

        If key is not found, d is returned if given, otherwise KeyError is
        raised.

        '''
        from xotl.tools.params import check_count
        count = len(args)
        check_count(count + 1, 1, 2, caller='pop')
        res = super().pop(key, Unset)
        if res is Unset:
            if count == 1:
                return args[0]
            else:
                raise KeyError(key)
        else:
            self._ranks.remove(key)
            return res

    def setdefault(self, key, default=None):
        '''D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D'''
        if key in self:
            return self[key]
        else:
            self[key] = default
            return default

    @recursive_repr()
    def __repr__(self):
        '''x.__repr__() <==> repr(x)'''
        aux = ', '.join('({!r}, {!r})'.format(k, self[k]) for k in self)
        if aux:
            aux = '[{}]'.format(aux)
        return '{}({})'.format(type(self).__name__, aux)

    def __reduce__(self):
        '''Return state information for pickling.'''
        cls = type(self)
        inst_dict = vars(self).copy()
        for k in vars(cls()):
            inst_dict.pop(k, None)
        return cls, (), inst_dict or None, None, iter(self.items())

    def copy(self):
        '''D.copy() -> a shallow copy of D.'''
        return type(self)(self)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        '''RD.fromkeys(S[, v]) -> New ranked dictionary with keys from S.

        If not specified, the value defaults to None.

        '''
        return cls((key, value) for key in iterable)


class OrderedSmartDict(SmartDictMixin, OrderedDict):
    '''A combination of the `OrderedDict` with the `SmartDictMixin`.

    .. warning:: Initializing with kwargs does not ensure any initial ordering,
                 since Python's keyword dict is not ordered. Use a list/tuple
                 of pairs instead.

    '''
    def __init__(*args, **kwds):
        '''Initialize an ordered dictionary.

        The signature is the same as regular dictionaries, but keyword
        arguments are not recommended because their insertion order is
        arbitrary.

        '''
        from xotl.tools.params import issue_9137
        self, args = issue_9137(args)
        super().__init__()
        self.update(*args, **kwds)


class MetaSet(type):
    '''Allow syntax sugar creating sets.

    This is pythonic syntax (stop limit is never included), for example::

        >>> from xotl.tools.future.collections import PascalSet as srange
        >>> [i for i in srange[1:4, 15, 20:23]]
        [1, 2, 3, 15, 20, 21, 22, 23]

    '''
    def __getitem__(cls, ranges):
        return cls(*ranges) if isinstance(ranges, tuple) else cls(ranges)


class PascalSet(metaclass=MetaSet):
    '''Collection of unique integer elements (implemented with intervals).

    ::

       PascalSet(*others) -> new set object

    .. versionadded:: 1.7.1

    '''
    __slots__ = ('_items',)

    def __init__(self, *others):
        '''Initialize self.

        :param others: Any number of integer or collection of integers that
               will be the set members.

        '''
        self._items = []    # a list of list of two elements
        self.update(*others)

    def __str__(self):
        def aux(s, e):
            if s == e:
                return str(s)
            elif s + 1 == e:
                return '%s, %s' % (s, e)
            else:
                return '%s..%s' % (s, e)

        l = self._items
        ranges = ((l[i], l[i + 1]) for i in range(0, len(l), 2))
        return str('{%s}' % ', '.join((aux(s, e) for (s, e) in ranges)))

    def __repr__(self):
        cls = type(self)
        cname = cls.__name__
        return str('%s([%s])' % (cname, ', '.join((str(i) for i in self))))

    def __iter__(self):
        l = self._items
        i, count = 0, len(l)
        while i < count:
            s, e = l[i], l[i + 1]
            while s <= e:
                yield s
                s += 1
            i += 2

    def __len__(self):
        res = 0
        l = self._items
        i, count = 0, len(l)
        while i < count:
            res += l[i + 1] - l[i] + 1
            i += 2
        return res

    def __nonzero__(self):
        return bool(self._items)
    __bool__ = __nonzero__

    def __contains__(self, other):
        '''True if set has an element ``other``, else False.'''
        return isinstance(other, int) and self._search(other)[0]

    def __hash__(self):
        '''Compute the hash value of a set.'''
        return Set._hash(self)

    def __eq__(self, other):
        '''Python 2 and 3 have several differences in operator definitions.

        For example::

          >>> from xotl.tools.future.collections import PascalSet
          >>> s1 = PascalSet[0:10]
          >>> assert s1 == set(s1)    # OK (True) in 2 and 3
          >>> assert set(s1) == s1    # OK in 3, fails in 2

        '''
        if isinstance(other, Set):
            ls, lo = len(self), len(other)
            if ls == lo:
                if isinstance(other, PascalSet):
                    return self._items == other._items
                else:
                    return self.count(other) == ls
            else:
                return False
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        if isinstance(other, Set):
            if other:
                return self.issuperset(other) and len(self) > len(other)
            else:
                return bool(self._items)
        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Set):
            if other:
                return self.issuperset(other)
            else:
                return bool(self._items)
        else:
            return NotImplemented

    def __lt__(self, other):
        if isinstance(other, Set):
            if other:
                return self.issubset(other) and len(self) < len(other)
            else:
                return not self._items
        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, Set):
            if other:
                return self.issubset(other)
            else:
                return not self._items
        else:
            return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Set):
            return self.difference(other)
        else:
            return NotImplemented

    def __isub__(self, other):
        if isinstance(other, Set):
            self.difference_update(other)
            return self
        else:
            return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, Set):
            return other - type(other)(self)
        else:
            return NotImplemented

    def __and__(self, other):
        if isinstance(other, Set):
            return self.intersection(other)
        else:
            return NotImplemented

    def __iand__(self, other):
        if isinstance(other, Set):
            self.intersection_update(other)
            return self
        else:
            return NotImplemented

    def __rand__(self, other):
        if isinstance(other, Set):
            return other & type(other)(self)
        else:
            return NotImplemented

    def __or__(self, other):
        if isinstance(other, Set):
            return self.union(other)
        else:
            return NotImplemented

    def __ior__(self, other):
        if isinstance(other, Set):
            self.update(other)
            return self
        else:
            return NotImplemented

    def __ror__(self, other):
        if isinstance(other, Set):
            return other | type(other)(self)
        else:
            return NotImplemented

    def __xor__(self, other):
        if isinstance(other, Set):
            return self.symmetric_difference(other)
        else:
            return NotImplemented

    def __ixor__(self, other):
        if isinstance(other, Set):
            self.symmetric_difference_update(other)
            return self
        else:
            return NotImplemented

    def __rxor__(self, other):
        if isinstance(other, Set):
            return other ^ type(other)(self)
        else:
            return NotImplemented

    def count(self, other):
        '''Number of occurrences of any member of other in this set.

        If other is an integer, return 1 if present, 0 if not.

        '''
        if isinstance(other, int):
            return 1 if other in self else 0
        else:
            return sum((i in self for i in other), 0)

    def add(self, other):
        '''Add an element to a set.

        This has no effect if the element is already present.

        '''
        self._insert(other)

    def union(self, *others):
        '''Return the union of sets as a new set.

        (i.e. all elements that are in either set.)

        '''
        res = self.copy()
        res.update(*others)
        return res

    def update(self, *others):
        '''Update a set with the union of itself and others.'''
        for other in others:
            if isinstance(other, PascalSet):
                l = other._items
                if self._items:
                    count = len(l)
                    i = 0
                    while i < count:
                        self._insert(l[i], l[i + 1])
                        i += 2
                else:
                    self._items = l[:]
            elif isinstance(other, int):
                self._insert(other)
            elif isinstance(other, Iterable):
                for i in other:
                    self._insert(i)
            elif isinstance(other, slice):
                start, stop, step = other.start, other.stop, other.step
                if step is None:
                    step = 1
                if step in (1, -1):
                    stop -= step
                    if step == -1:
                        start, stop = stop, start
                    self._insert(start, stop)
                else:
                    for i in range(start, stop, step):
                        self._insert(i)
            else:
                raise self._invalid_value(other)

    def intersection(self, *others):
        '''Return the intersection of two or more sets as a new set.

        (i.e. elements that are common to all of the sets.)

        '''
        res = self.copy()
        res.intersection_update(*others)
        return res

    def intersection_update(self, *others):
        '''Update a set with the intersection of itself and another.'''
        l = self._items
        oi, count = 0, len(others)
        while l and oi < count:
            other = others[oi]
            if not isinstance(other, PascalSet):
                # safe mode for intersection
                other = PascalSet(i for i in other if isinstance(i, int))
            o = other._items
            if o:
                sl, el = l[0], l[-1]
                so, eo = o[0], o[-1]
                if sl < so:
                    self._remove(sl, so - 1)
                if eo < el:
                    self._remove(eo + 1, el)
                i = 2
                while l and i < len(o):
                    s, e = o[i - 1] + 1, o[i] - 1
                    if s <= e:
                        self._remove(s, e)
                    i += 2
            else:
                del l[:]
            oi += 1

    def difference(self, *others):
        '''Return the difference of two or more sets as a new set.

        (i.e. all elements that are in this set but not the others.)

        '''
        res = self.copy()
        res.difference_update(*others)
        return res

    def difference_update(self, *others):
        '''Remove all elements of another set from this set.'''
        for other in others:
            if isinstance(other, PascalSet):
                l = other._items
                count = len(l)
                i = 0
                while i < count:
                    self._remove(l[i], l[i + 1])
                    i += 2
            else:
                for i in other:
                    if isinstance(i, int):
                        self._remove(i)

    def symmetric_difference(self, other):
        '''Return the symmetric difference of two sets as a new set.

        (i.e. all elements that are in exactly one of the sets.)

        '''
        res = self.copy()
        res.symmetric_difference_update(other)
        return res

    def symmetric_difference_update(self, other):
        'Update a set with the symmetric difference of itself and another.'
        if not isinstance(other, PascalSet):
            other = PascalSet(other)
        if self:
            if other:
                # TODO: Implement more efficiently
                aux = other - self
                self -= other
                self |= aux
        else:
            self._items = other._items[:]

    def discard(self, other):
        '''Remove an element from a set if it is a member.

        If the element is not a member, do nothing.

        '''
        self._remove(other)

    def remove(self, other):
        '''Remove an element from a set; it must be a member.

        If the element is not a member, raise a KeyError.

        '''
        if other in self:
            self._remove(other)
        else:
            raise KeyError('"%s" is not a member!' % other)

    def pop(self):
        '''Remove and return an arbitrary set element.

        Raises KeyError if the set is empty.

        '''
        l = self._items
        if l:
            res = l[0]
            if l[0] < l[1]:
                l[0] += 1
            else:
                del l[0:2]
            return res
        else:
            raise KeyError('pop from an empty set!')

    def clear(self):
        '''Remove all elements from this set.'''
        self._items = []

    def copy(self):
        '''Return a shallow copy of a set.'''
        return type(self)(self)

    def isdisjoint(self, other):
        '''Return True if two sets have a null intersection.'''
        if isinstance(other, PascalSet):
            if self and other:
                l, o = self._items, other._items
                i, lcount, ocount = 0, len(l), len(o)
                maybe = True
                while maybe and i < lcount:
                    found, idx = other._search(l[i])
                    if idx == ocount:    # exhausted
                        # assert not found
                        i = lcount
                    elif found or l[i + 1] >= o[idx]:
                        maybe = False
                    else:
                        i += 2
                return maybe
            else:
                return True
        else:
            return not any(i in self for i in other)

    def issubset(self, other):
        '''Report whether another set contains this set.'''
        ls = len(self)
        if isinstance(other, PascalSet):
            if self:
                if ls > len(other):  # Fast check for obvious cases
                    return False
                else:
                    l, o = self._items, other._items
                    i, lcount = 0, len(l)
                    maybe = True
                    while maybe and i < lcount:
                        found, idx = other._search(l[i])
                        if found and l[i + 1] <= o[idx + 1]:
                            i += 2
                        else:
                            maybe = False
                    return maybe
            else:
                return True
        elif isinstance(other, Sized) and ls > len(other):
            # Fast check for obvious cases
            return False
        elif isinstance(other, Container):
            aux = next((i for i in self if i not in other), Unset)
            return aux is Unset
        else:
            # Generator cases
            from operator import add
            from functools import reduce
            lo = reduce(add, (i in self for i in other), 0)
            return lo == ls

    def issuperset(self, other):
        '''Report whether this set contains another set.'''
        ls = len(self)
        if isinstance(other, PascalSet):
            if other:
                if ls < len(other):  # Fast check for obvious cases
                    return False
                else:
                    l, o = self._items, other._items
                    i, ocount = 0, len(o)
                    maybe = True
                    while maybe and i < ocount:
                        found, idx = self._search(o[i])
                        if found and o[i + 1] <= l[idx + 1]:
                            i += 2
                        else:
                            maybe = False
                    return maybe
            else:
                return True
        elif isinstance(other, Sized) and ls < len(other):
            # Fast check for obvious cases
            return False
        else:
            aux = next((i for i in other if i not in self), Unset)
            return aux is Unset

    def _search(self, other):
        '''Search the pair where ``other`` is placed.

        Return a duple :``(if found or not, index)``.

        '''
        if isinstance(other, int):
            l = self._items
            start, end = 0, len(l)
            res, pivot = False, 2 * (end // 4)
            while not res and start < end:
                s, e = l[pivot], l[pivot + 1]
                if other < s:
                    end = pivot
                elif other > e:
                    start = pivot + 2
                else:
                    res = True
                pivot = start + 2 * ((end - start) // 4)
            return res, pivot
        else:
            raise self._invalid_value(other)

    def _insert(self, start, end=None):
        '''Insert an interval of integers.'''
        if not end:
            end = start
        assert start <= end
        l = self._items
        count = len(l)
        found, idx = self._search(start)
        if not found:
            if idx > 0 and start == l[idx - 1] + 1:
                found = True
                idx -= 2
                l[idx + 1] = start
                if idx < count - 2 and end == l[idx + 2] - 1:
                    end = l[idx + 3]
            elif idx < count and end >= l[idx] - 1:
                found = True
                l[idx] = start
        if found:
            while end > l[idx + 1]:
                if idx < count - 2 and end >= l[idx + 2] - 1:
                    if end <= l[idx + 3]:
                        l[idx + 1] = l[idx + 3]
                    del l[idx + 2:idx + 4]
                    count -= 2
                else:
                    l[idx + 1] = end
        else:
            if idx < count:
                l.insert(idx, start)
                l.insert(idx + 1, end)
            else:
                l.extend((start, end))
            count += 2

    def _remove(self, start, end=None):
        '''Remove an interval of integers.'''
        if not end:
            end = start
        assert start <= end
        l = self._items
        sfound, sidx = self._search(start)
        efound, eidx = self._search(end)
        if sfound and efound and sidx == eidx:
            first = l[sidx] < start
            last = l[eidx + 1] > end
            if first and last:
                l.insert(eidx + 1, end + 1)
                l.insert(sidx + 1, start - 1)
            elif first:
                l[sidx + 1] = start - 1
            elif last:
                l[eidx] = end + 1
            else:
                del l[sidx:eidx + 2]
        else:
            if sfound and l[sidx] < start:
                l[sidx + 1] = start - 1
                sidx += 2
            if efound:
                if l[eidx + 1] > end:
                    l[eidx] = end + 1
                else:
                    eidx += 2
            if sidx < eidx:
                del l[sidx:eidx]

    def _invalid_value(self, value):
        cls_name = type(self).__name__
        vname = type(value).__name__
        msg = ('Unsupported type for  value "%s" of type "%s" for a "%s", '
               'must be an integer!')
        return TypeError(msg % (value, vname, cls_name))

    @classmethod
    def _prime_numbers_until(cls, limit):
        '''This is totally a funny test method.'''
        res = cls[2:limit]
        for i in range(2, limit // 2 + 1):
            if i in res:
                aux = i + i
                while aux < limit:
                    if aux in res:
                        res.remove(aux)
                    aux += i
        return res


MutableSet.register(PascalSet)


class BitPascalSet(metaclass=MetaSet):
    '''Collection of unique integer elements (implemented with bit-wise sets).

    ::

        BitPascalSet(*others) -> new bit-set object

    .. versionadded:: 1.7.1

    '''
    __slots__ = ('_items',)
    _bit_length = 62    # How many values are stored in each item

    def __init__(self, *others):
        '''Initialize self.

        :param others: Any number of integer or collection of integers that
               will be the set members.

        In this case `_items` is a dictionary with keys containing number
        division seeds and values bit-wise integers (each bit is the division
        modulus position).

        '''
        self._items = {}
        self.update(*others)

    def __str__(self):
        if self:
            return str(PascalSet(self))
        else:
            cname = type(self).__name__
            return str('%s([])') % cname

    def __repr__(self):
        cname = type(self).__name__
        res = str(', ').join(str(i) for i in self)
        return str('%s([%s])') % (cname, res)

    def __iter__(self):
        bl = self._bit_length
        sm = self._items
        for k in sorted(sm):
            v = sm[k]
            base = k * bl
            i = 0
            ref = 1
            while i < bl:
                if ref & v:
                    yield base + i
                ref <<= 1
                i += 1

    def __len__(self):
        return sum((1 for i in self), 0)

    def __nonzero__(self):
        return bool(self._items)
    __bool__ = __nonzero__

    def __contains__(self, other):
        '''True if this bit-set has the element ``other``, else False.'''
        res = self._search(other)
        if res:
            k, ref, v = res
            return bool(v & (1 << ref))
        else:
            return False

    def __hash__(self):
        '''Compute the hash value of a set.'''
        return Set._hash(self)

    def __eq__(self, other):
        '''Python 2 and 3 have several differences in operator definitions.

        For example::

          >>> from xotl.tools.future.collections import BitPascalSet
          >>> s1 = BitPascalSet[0:10]
          >>> assert s1 == set(s1)    # OK (True) in 2 and 3
          >>> assert set(s1) == s1    # OK in 3, fails in 2

        '''
        if isinstance(other, Set):
            if isinstance(other, BitPascalSet):
                return self._items == other._items
            else:
                ls, lo = len(self), len(other)
                return ls == lo == self.count(other)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        if isinstance(other, Set):
            if other:
                return self.issuperset(other) and len(self) > len(other)
            else:
                return bool(self._items)
        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Set):
            return self.issuperset(other) if other else bool(self._items)
        else:
            return NotImplemented

    def __lt__(self, other):
        if isinstance(other, Set):
            if other:
                return self.issubset(other) and len(self) < len(other)
            else:
                return not self._items
        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, Set):
            return self.issubset(other) if other else not self._items
        else:
            return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Set):
            return self.difference(other)
        else:
            return NotImplemented

    def __isub__(self, other):
        if isinstance(other, Set):
            self.difference_update(other)
            return self
        else:
            return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, Set):
            return other - type(other)(self)
        else:
            return NotImplemented

    def __and__(self, other):
        if isinstance(other, Set):
            return self.intersection(other)
        else:
            return NotImplemented

    def __iand__(self, other):
        if isinstance(other, Set):
            self.intersection_update(other)
            return self
        else:
            return NotImplemented

    def __rand__(self, other):
        if isinstance(other, Set):
            return other & type(other)(self)
        else:
            return NotImplemented

    def __or__(self, other):
        if isinstance(other, Set):
            return self.union(other)
        else:
            return NotImplemented

    def __ior__(self, other):
        if isinstance(other, Set):
            self.update(other)
            return self
        else:
            return NotImplemented

    def __ror__(self, other):
        if isinstance(other, Set):
            return other | type(other)(self)
        else:
            return NotImplemented

    def __xor__(self, other):
        if isinstance(other, Set):
            return self.symmetric_difference(other)
        else:
            return NotImplemented

    def __ixor__(self, other):
        if isinstance(other, Set):
            self.symmetric_difference_update(other)
            return self
        else:
            return NotImplemented

    def __rxor__(self, other):
        if isinstance(other, Set):
            return other ^ type(other)(self)
        else:
            return NotImplemented

    def count(self, other):
        '''Number of occurrences of any member of other in this set.

        If other is an integer, return 1 if present, 0 if not.

        '''
        if isinstance(other, int):
            return 1 if other in self else 0
        else:
            return sum((i in self for i in other), 0)

    def add(self, other):
        '''Add an element to a bit-set.

        This has no effect if the element is already present.

        '''
        self._insert(other)

    def union(self, *others):
        '''Return the union of bit-sets as a new set.

        (i.e. all elements that are in either set.)

        '''
        res = self.copy()
        res.update(*others)
        return res

    def update(self, *others):
        '''Update a bit-set with the union of itself and others.'''
        for other in others:
            if isinstance(other, BitPascalSet):
                sm = self._items
                om = other._items
                for k, v in safe_dict_iter(om).items():
                    if k in sm:
                        sm[k] |= v
                    else:
                        sm[k] = v
            elif isinstance(other, int):
                self._insert(other)
            elif isinstance(other, Iterable):
                for i in other:
                    self._insert(i)
            elif isinstance(other, slice):
                start, stop, step = other.start, other.stop, other.step
                if step is None:
                    step = 1
                for i in range(start, stop, step):
                    self._insert(i)
            else:
                raise self._invalid_value(other)

    def intersection(self, *others):
        '''Return the intersection of two or more bit-sets as a new set.

        (i.e. elements that are common to all of the sets.)

        '''
        res = self.copy()
        res.intersection_update(*others)
        return res

    def intersection_update(self, *others):
        '''Update a bit-set with the intersection of itself and another.'''
        sm = self._items
        oi, count = 0, len(others)
        while sm and oi < count:
            other = others[oi]
            if not isinstance(other, BitPascalSet):
                # safe mode for intersection
                other = PascalSet(i for i in other if isinstance(i, int))
            om = other._items
            for k, v in safe_dict_iter(sm).items():
                v &= om.get(k, 0)
                if v:
                    sm[k] = v
                else:
                    del sm[k]
            oi += 1

    def difference(self, *others):
        '''Return the difference of two or more bit-sets as a new set.

        (i.e. all elements that are in this set but not the others.)

        '''
        res = self.copy()
        res.difference_update(*others)
        return res

    def difference_update(self, *others):
        '''Remove all elements of another bit-set from this set.'''
        for other in others:
            if isinstance(other, BitPascalSet):
                sm = self._items
                om = other._items
                for k, v in safe_dict_iter(om).items():
                    if k in sm:
                        v = sm[k] & ~v
                        if v:
                            sm[k] = v
                        else:
                            del sm[k]
            else:
                for i in other:
                    if isinstance(i, int):
                        self._remove(i)

    def symmetric_difference(self, other):
        '''Return the symmetric difference of two bit-sets as a new set.

        (i.e. all elements that are in exactly one of the sets.)

        '''
        res = self.copy()
        res.symmetric_difference_update(other)
        return res

    def symmetric_difference_update(self, other):
        'Update a bit-set with the symmetric difference of itself and another.'
        if not isinstance(other, BitPascalSet):
            other = BitPascalSet(other)
        if self:
            if other:
                # TODO: Implement more efficiently
                aux = other - self
                self -= other
                self |= aux
        else:
            self._items = other._items[:]

    def discard(self, other):
        '''Remove an element from a bit-set if it is a member.

        If the element is not a member, do nothing.

        '''
        self._remove(other)

    def remove(self, other):
        '''Remove an element from a bit-set; it must be a member.

        If the element is not a member, raise a KeyError.

        '''
        self._remove(other, fail=True)

    def pop(self):
        '''Remove and return an arbitrary bit-set element.

        Raises KeyError if the set is empty.

        '''
        sm = self._items
        if sm:
            bl = self._bit_length
            k, v = next(sm.items())
            assert v
            base = k * bl
            i = 0
            ref = 1
            res = None
            while res is None:
                if ref & v:
                    res = base + i
                else:
                    ref <<= 1
                    i += 1
            v &= ~ref
            if v:
                sm[k] = v
            else:
                del sm[k]
            return res
        else:
            raise KeyError('pop from an empty set!')

    def clear(self):
        '''Remove all elements from this bit-set.'''
        self._items = {}

    def copy(self):
        '''Return a shallow copy of a set.'''
        return type(self)(self)

    def isdisjoint(self, other):
        '''Return True if two bit-sets have a null intersection.'''
        if isinstance(other, BitPascalSet):
            sm, om = self._items, other._items
            if sm and om:
                return all(sm.get(k, 0) & v == 0 for k, v in om.items())
            else:
                return True
        else:
            return not any(i in self for i in other)

    def issubset(self, other):
        '''Report whether another set contains this bit-set.'''
        if isinstance(other, BitPascalSet):
            sm, om = self._items, other._items
            if sm:
                return all(om.get(k, 0) & v == v for k, v in sm.items())
            else:
                return True
        elif isinstance(other, Container):
            return not any(i not in other for i in self)
        else:
            # Generator cases
            return sum((i in self for i in other), 0) == len(self)

    def issuperset(self, other):
        '''Report whether this bit set contains another set.'''
        if isinstance(other, BitPascalSet):
            sm, om = self._items, other._items
            if om:
                return all(sm.get(k, 0) & v == v for k, v in om.items())
            else:
                return True
        else:
            return not any(i not in self for i in other)

    def _search(self, other):
        '''Search the bit-wise value where ``other`` could be placed.

        Return a duple :``(seed, bits to shift left)``.

        '''
        if isinstance(other, int):
            sm = self._items
            bl = self._bit_length
            k, ref = divmod(other, bl)
            return k, ref, sm.get(k, 0)
        else:
            return None

    def _insert(self, other):
        '''Add a member in this bit-set.'''
        aux = self._search(other)
        if aux:
            k, ref, v = aux
            self._items[k] = v | (1 << ref)
        else:
            raise self._invalid_value(other)

    def _remove(self, other, fail=False):
        '''Remove an interval of integers from this bit-set.'''
        aux = self._search(other)
        ok = False
        if aux:
            k, ref, v = aux
            if v:
                aux = v & ~(1 << ref)
                if v != aux:
                    ok = True
                    sm = self._items
                    if aux:
                        sm[k] = aux
                    else:
                        del sm[k]
        if not ok and fail:
            raise KeyError('"%s" is not a member!' % other)

    def _invalid_value(self, value):
        cls_name = type(self).__name__
        vname = type(value).__name__
        msg = ('Unsupported type for  value "%s" of type "%s" for a "%s", '
               'must be an integer!')
        return TypeError(msg % (value, vname, cls_name))

    @classmethod
    def _prime_numbers_until(cls, limit):
        '''This is totally a funny test method.'''
        res = cls[2:limit]
        for i in range(2, limit // 2 + 1):
            if i in res:
                aux = i + i
                while aux < limit:
                    if aux in res:
                        res.remove(aux)
                    aux += i
        return res


MutableSet.register(BitPascalSet)


# Smart Tools

def pair(arg):
    '''Check if `arg` is a pair (two elements tuple or list).'''
    return arg if isinstance(arg, (tuple, list)) and len(arg) == 2 else None


def smart_iter_items(*args, **kwds):
    '''Unify iteration over (key, value) pairs.

    If a pair of pairs is found, e.g. ``[(1, 2), (3, 4)]``, instead of
    yielding one pair (the outermost), inner two pairs takes precedence.

    '''
    for arg in args:
        if isinstance(arg, Mapping):
            for key in arg:
                yield key, arg[key]
        elif hasattr(arg, 'keys') and hasattr(arg, '__getitem__'):
            # Custom mappings
            for key in arg.keys():
                yield key, arg[key]
        elif pair(arg) and not (pair(arg[0]) and pair(arg[1])):
            yield arg
        else:
            for item in arg:
                if pair(item):
                    yield item
                else:
                    msg = "'{}' object '{}' is not a pair"
                    raise TypeError(msg.format(type(item).__name__, item))
    for key in kwds:
        yield key, kwds[key]


# get rid of unused global variables
del deprecated, recursive_repr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''A context manager for execution context flags.'''

from xotl.tools.tasking import local
from xotl.tools.future.collections import StackedDict, Mapping

__all__ = ('Context', 'context', 'NullContext')


class LocalData(local):
    '''Thread-local data for contexts.'''
    def __init__(self):
        super().__init__()
        self.contexts = {}


_data = LocalData()


class MetaContext(type(StackedDict)):  # type: ignore
    def __len__(self):
        return len(_data.contexts)

    def __iter__(self):
        return iter(_data.contexts)

    def __getitem__(self, name):
        return _data.contexts.get(name, _null_context)

    def __contains__(self, name):
        '''Basic support for the 'A in context' idiom.'''
        return bool(self[name])


class Context(StackedDict, metaclass=MetaContext):
    '''An execution context manager with parameters (or flags).

    Use as::

        >>> SOME_CONTEXT = object()
        >>> from xotl.tools.context import context
        >>> with context(SOME_CONTEXT):
        ...     if context[SOME_CONTEXT]:
        ...         print('In context SOME_CONTEXT')
        In context SOME_CONTEXT

    Note the difference creating the context and checking it: for entering a
    context you should use ``context(name)`` for testing whether some piece of
    code is being executed inside a context you should use ``context[name]``;
    you may also use the syntax `name in context`.

    When an existing context is re-enter, the former one is reused.
    Nevertheless, the data stored in each context is local to each level.

    For example::

        >>> with context('A', b=1) as a1:
        ...   with context('A', b=2) as a2:
        ...       print(a1 is a2)
        ...       print(a2['b'])
        ...   print(a1['b'])
        True
        2
        1

    For data access, a mapping interface is provided for all contexts. If a
    data slot is deleted at some level, upper level is used to read
    values. Each new written value is stored in current level without
    affecting upper levels.

    For example::

        >>> with context('A', b=1) as a1:
        ...   with context('A', b=2) as a2:
        ...       del a2['b']
        ...       print(a2['b'])
        1

    It is an error to *reuse* a context directly like in::

        >>> with context('A', b=1) as a1:   # doctest: +ELLIPSIS
        ...   with a1:
        ...       pass
        Traceback (most recent call last):
        ...
        RuntimeError: Entering the same context level twice! ...

    '''
    __slots__ = ('name', 'count', )

    def __new__(cls, name, **data):
        self = cls[name]
        if not self:     # if self is _null_context:
            self = super().__new__(cls)
            super(Context, self).__init__()
            self.name = name
            self.count = 0
            # TODO: Redefine all event management
        return self(**data)

    @classmethod
    def from_dicts(cls, ctx, overrides=None, defaults=None):
        '''Creates a context introducing both defaults and overrides.

        This combines both the standard constructor and `from_defaults`:meth:.

        If the same key appears in both `overrides` and `defaults`, ignore the
        default.

        '''
        if not overrides:
            overrides = {}
        if not defaults:
            defaults = {}
        current = cls[ctx]
        current_attrs = dict(current) if current else {}
        attrs = dict(defaults, **current_attrs)
        attrs.update(overrides)
        return cls(ctx, **attrs)

    @classmethod
    def from_defaults(cls, ctx, **defaults):
        '''Creates context `ctx` introducing only new keys given in `defaults`.

        The normal behavior when you enter a new level in the context is to
        override the values with the new one.

        Example:

           >>> with context.from_defaults('A', a=1):
           ...    with context.from_defaults('A', a=2, b=1) as c:
           ...        assert c['a'] == 1

        '''
        return cls.from_dicts(ctx, defaults=defaults)

    def __init__(self, *args, **kwargs):
        '''Must be defined empty for `__new__` parameters compatibility.

        Using generic parameters definition allow any redefinition of this
        class can use this `__init__`.

        '''

    def __call__(self, **data):
        '''Allow re-enter in a new level to an already assigned context.'''
        self.push_level(**data)
        return self

    def __nonzero__(self):
        return bool(self.count)
    __bool__ = __nonzero__

    def __enter__(self):
        if self.count == 0:
            _data.contexts[self.name] = self
        if self.count + 1 == self.level:
            self.count += 1
            return self
        else:
            msg = 'Entering the same context level twice! -- c(%s, %d, %d)'
            raise RuntimeError(msg % (self.name, self.count, self.level))

    def __exit__(self, exc_type, exc_value, traceback):
        self.count -= 1
        if self.count == 0:
            del _data.contexts[self.name]
        self.pop_level()
        return False


# A simple alias for Context
context = Context


class NullContext(Mapping):
    '''Singleton context to be used (returned) as default when no one is
    defined.

    '''

    __slots__ = ()

    instance = None
    name = ''

    def __new__(cls):
        if cls.instance is None:
            cls.instance = super().__new__(cls)
        return cls.instance

    def __len__(self):
        return 0

    def __iter__(self):
        return iter(())

    def __getitem__(self, key):
        raise KeyError(key)

    def __nonzero__(self):
        return False
    __bool__ = __nonzero__

    def __enter__(self):
        return _null_context

    def __exit__(self, exc_type, exc_value, traceback):
        return False

    def get(self, name, default=None):
        return default

    @property
    def level(self):
        return 0


_null_context = NullContext()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''The standard `physical quantities`_.

.. _physical quantities: \
https://en.wikipedia.org/wiki/International_System_of_Quantities#Base_quantities

'''

from .meta import (
    Dimension,
    UNIT,
)


def kilo(v):
    return 1000 * v


def centi(v):
    return v / 100


def milli(v):
    return v / 1000


def micro(v):
    return v / 1000000


def nano(v):
    return v / (10**9)


@Dimension.new()
class Length:
    metre = UNIT
    kilometer = km = kilo(metre)
    centimeter = cm = centi(metre)
    millimeter = mm = milli(metre)
    nanometer = nm = nano(metre)


metre = m = Length.m = Length.metre
L = Length


@Dimension.new
class Time:
    second = UNIT
    millisecond = ms = milli(second)
    nanosecond = ns = nano(second)
    minute = second * 60
    hour = minute * 60


second = s = Time.s = Time.second
T = Time


@Dimension.new(unit_aliases=('kg', ))
class Mass:
    kilogram = UNIT
    gram = kilogram / 1000


kilogram = kg = Mass.kg
M = Mass


@Dimension.new(unit_aliases='A')
class ElectricCurrent:
    ampere = UNIT
    milliampere = milli(ampere)


A = ampere = ElectricCurrent.A
I = ElectricCurrent


@Dimension.new(unit_aliases='K')
class Temperature:
    kelvin = UNIT

    @classmethod
    def from_celcius(cls, val):
        'Convert `val` ºC to K'
        return (val + 273.15) * cls.kelvin

    @classmethod
    def from_fahrenheit(cls, val):
        'Convert `val` ºF to K'
        return (val + 459.67) * (5 / 9) * cls.kelvin


K = kelvin = Temperature.K
O = Temperature   # The actual symbol would be the capital letter Theta: Θ


@Dimension.new(unit_alias='mol')
class Substance:
    mole = UNIT


mole = mol = Substance.mol
N = Substance


@Dimension.new
class Luminosity:
    candela = UNIT


J = Luminosity


# Derived quantities
Area = L**2
Volume = L**3
Volume.metre_cubic = Volume._unit_
Volume._unitname_ = 'metre_cubic'

Frequency = T**-1
Frequency.Hz = Frequency._unit_

Force = L * M / T**2
assert hasattr(Force, 'metre_kilogram_per_second_squared')
assert Force == L * M * T**-2
Force.Newton = Force.N = Force._unit_

Presure = M / L / T**2
assert hasattr(Presure, 'kilogram_per_metre_per_second_squared')
assert Presure == L**-1 * M * T**-2, 'as defined in Wikipedia'
Presure.Pascal = Presure.Pa = Presure._unit_

Velocity = L / T
Acceleration = L / T**2
